"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var isclose_scalers_1 = require("./isclose.scalers");
/**
 * Check if a is approximately equal to b,
 * or for arrays,
 * if each element of a is approximately equal to each element of b.
 * @param {number|Array<number>} a First value(s) to compare.
 * @param {number|Array<number>} b Second value(s) to compare.
 * @param {number} rtol The relative tolerance parameter, default = 1e-09.
 * @param {number} atol The absolute tolerance parameter, default = 0.0.
 * @param {boolean} equalNaN Whether to compare NaNs as equal, default = false.
 *                  If true, when a is NaN and b is NaN they will be considered equal.
 * @param {IsCloseScaler} rtolScaler
 *                        Determine how the relative difference will be scaled using a function rtolScaler(a,b).
 * @returns {boolean|Array<boolean>}
 *          For singletons:
 *          Returns true when both a and b are considered to be close,
 *          that is:
 *            abs(a - b) <= max(rtol * rtolScaler(a,b), atol)
 *          For Arrays:
 *          Returns an array of booleans corresponding to the result of calling isClose() on the individual values.
 */
function isClose(a, b, rtol, atol, equalNaN, rtolScaler) {
    if (rtol === void 0) { rtol = 1e-09; }
    if (atol === void 0) { atol = 0.0; }
    if (equalNaN === void 0) { equalNaN = false; }
    if (rtolScaler === void 0) { rtolScaler = isclose_scalers_1.IsCloseScalers.maxAbsAOrB; }
    if (typeof a === 'number') {
        var result = false;
        if (equalNaN && isNaN(a) && isNaN(b)) {
            // NaNs can, optionally, be close to other NaNs
            result = true;
        }
        else if (a === b) {
            // Same values are always close
            result = true;
        }
        else if (a === Infinity || a === -Infinity || b === Infinity || b === -Infinity) {
            // Infinities are never close to other values except themselves
            result = false;
        }
        else {
            var adiff = Math.abs(a - b);
            if (atol >= adiff) {
                result = true;
            }
            else {
                var rdiff = adiff / rtolScaler(a, b);
                if (rtol >= rdiff) {
                    result = true;
                }
            }
        }
        return result;
    }
    else if (a instanceof Array) {
        var result = new Array(Math.max(a.length, b.length));
        var min = Math.min(a.length, b.length);
        for (var i = 0; i < min; i++) {
            result[i] = isClose(a[i], b[i], rtol, atol, equalNaN, rtolScaler);
        }
        return result;
    }
    throw new TypeError('a and b must both be either numbers or arrays of numbers.');
}
exports.isClose = isClose;
